from x_transformers.x_transformers import *
from x_transformers import Decoder
from torch.nn import CrossEntropyLoss

class LucidLM2HF(nn.Module):
    def __init__(self, config, addl_decoder_config={}):
        r'''
        '''
        super().__init__()
        self.config = config # Before we turn int into a dict, save it
        config = config.to_dict() # Make nicer if possible (currently done so that we have `get`)

        attn_layers = Decoder(dim=config['n_embd'], depth=config['n_layer'], heads=config['n_head'], **addl_decoder_config)

        self.token_emb = nn.Embedding(config['vocab_size'], config['n_embd'])
        self.pos_emb = AbsolutePositionalEmbedding(config['n_embd'], config['n_positions']) if (config.get('use_pos_emb', True) \
                                                   and not attn_layers.has_pos_emb) else always(0)
        self.emb_dropout = nn.Dropout(config.get('emb_dropout', 0.))

        #self.project_emb = nn.Linear(config['n_embd'], config['n_embd']) if config['n_embd'] != config.get('n_embd_proj', config['n_embd']) \
        #                    else nn.Identity()
        self.project_emb = nn.Identity()

        self.attn_layers = attn_layers
        self.norm = nn.LayerNorm(config['n_embd'])

        self.init_()

        self.to_logits = nn.Linear(config['n_embd'], config['vocab_size']) if not config.get('tie_embedding', False) \
                                   else lambda t: t @ self.token_emb.weight.t()

        # memory tokens (like [cls]) from Memory Transformers paper
        num_memory_tokens = config.get('num_memory_tokens', 0)
        self.num_memory_tokens = num_memory_tokens
        if num_memory_tokens > 0:
            self.memory_tokens = nn.Parameter(torch.randn(num_memory_tokens, config['n_embd']))

            # let funnel encoder know number of memory tokens, if specified
            # TODO: think of a cleaner solution
            if hasattr(attn_layers, 'num_memory_tokens'):
                attn_layers.num_memory_tokens = num_memory_tokens

        # Use this to pass old mems to a new run
        self.mems = None


    def init_(self):
        nn.init.normal_(self.token_emb.weight, std = 0.02)


    def forward(self, input_ids, attention_mask=None, labels=None, **kwargs):
        r''' Runs the forward pass.

        Args:
            input_ids
            attention_mask
            labels
        '''
        config = self.config.to_dict()
        output = {} # Return a dictionary, easier than to find positions at which something has to be
        x = input_ids # Lazy

        b, n, device, num_mem = *x.shape, x.device, self.num_memory_tokens
        x = self.token_emb(x)
        x += self.pos_emb(x)
        x = self.emb_dropout(x)

        x = self.project_emb(x)

        if num_mem > 0:
            mem = repeat(self.memory_tokens, 'n d -> b n d', b = b)
            x = torch.cat((mem, x), dim = 1)

            # auto-handle masking after appending memory tokens
            if attention_mask is not None:
                attention_mask = F.pad(attention_mask, (num_mem, 0), value = True)


        x, intermediates = self.attn_layers(x, mask=attention_mask, mems=self.mems, return_hiddens=True, **kwargs)
        x = self.norm(x)

        mem, x = x[:, :num_mem], x[:, num_mem:]

        lm_logits = self.to_logits(x)
        output['logits'] = lm_logits

        if labels is not None:
            # Shift labels/ids to deal with ids==labels
            shift_logits = lm_logits[..., :-1, :].contiguous()
            shift_labels = labels[..., 1:].contiguous()
            # Flatten the tokens
            loss_fct = CrossEntropyLoss()
            loss = loss_fct(shift_logits.view(-1, shift_logits.size(-1)), shift_labels.view(-1))
            output['loss'] = loss
        if config.get('return_mems', False):
            hiddens = intermediates.hiddens
            new_mems = list(map(lambda pair: torch.cat(pair, dim = -2), zip(self.mems, hiddens))) if exists(self.mems) else hiddens
            new_mems = list(map(lambda t: t[..., -config.get('max_mem_len', 0.):, :].detach(), new_mems))
            output['new_mems'] = new_mems
            self.mems = new_mems # If I understood correctly how this works
        if config.get('return_attn', False):
            attn_maps = list(map(lambda t: t.post_softmax_attn, intermediates.attn_intermediates))
            output['attn_maps'] = attn_maps

        return output



def filter_by_count(dataset, min_count=5, min_length=5):
    r''' Filters tokens of a dataset and leaves only the ones with frequencey >= min_count

    Args:
        dataset
        min_count
        min_length:
            Examples below will be removed, in other words patients with less than min_length concepts
    '''
    token_cnt = {}
    for stream in dataset['stream']:
        for tkn in stream:
            token_cnt[tkn] = token_cnt.get(tkn, 0) + 1

    dataset = dataset.map(function=lambda example: {'stream': [token for token in example['stream'] if token_cnt[token] >= min_count]},
                      load_from_cache_file=False)

    if min_length > 0:
        dataset = dataset.filter(function=lambda example: len(example['stream']) >=  min_length)

    return dataset



from typing import Any, Callable, Dict, List, NewType, Optional, Tuple, Union
import torch

InputDataClass = NewType("InputDataClass", Any)

class CollataAndPad(object):
    r''' Arrange the data into the right format + add padding or trim where necessary.

    Args:
        max_len (`int`, `optional`, defaults to -1):
            Upper bound for sequence length. If it is -1 means that it will be
            calculated for each bach and set to the max length without upper limits.
        pad_id (`int`, `optional`, defaults to 0):
            What ID will be used to pad the inputs to max_len
        shift_labels:
            If True labels = input_ids[1:] + [pad_id], else labels = input_ids, this will also
                remove the last element from each sample in input_ids.
        remap_names:
            If a dictionary the stanard output names (input_ids, labels, attention_mask) will be maped
            to whatever is in the dict.
    '''
    def __init__(self, max_len=-1, pad_id=0, shift_labels=False, remap_names=None):
        self.max_len = max_len
        self.pad_id = pad_id
        self.shift_labels = shift_labels
        self.remap_names = remap_names


    def __call__(self, features: List[InputDataClass]) -> Dict[str, torch.Tensor]:
        batch = {}
        if self.max_len == -1:
            max_len = max([len(f['input_ids']) for f in features])
        else:
            max_len = min(self.max_len, max([len(f['input_ids']) for f in features]))

        if self.shift_labels and not 'labels' in features[0]:
            # Recalculate max_len
            max_len = min(max_len, max([len(f['input_ids'][0:-1]) for f in features]))

            # Labels do not exist and we should shift
            batch['labels'] = torch.tensor([f['input_ids'][1:max_len+1] + [-100] * max(0, max_len - len(f['input_ids']) + 1)
                                          for f in features], dtype=torch.long)

            for f in features:
                f['input_ids'] = f['input_ids'][0:-1]

        elif 'labels' in features[0]:
            # Labels already exist, just pad them with -100
            batch['labels'] = torch.tensor([f['labels'] for f in features], dtype=torch.long)

        batch['input_ids'] = torch.tensor([f['input_ids'][0:max_len] + [self.pad_id] * max(0, max_len - len(f['input_ids']))
                                          for f in features], dtype=torch.long)
        batch['attention_mask'] = batch['input_ids'] != self.pad_id

        if not self.shift_labels and not 'labels' in batch:
            # If we did not shift labels just clone
            batch['labels'] = batch['input_ids'].clone()
            batch['labels'][batch['labels'] == self.pad_id] = -100

        if self.remap_names is not None:
            batch = {self.remap_names.get(k, k):v for k,v in batch.items()}

        return batch




import numpy as np
import pandas as pd


def precision(predictions, label_ids, id2tkn, id2type, token_types={'cui'}, prediction_scope='one', shifted_labels=False,
        predictions_are_scores=True, old_data=None, topk=1, start=0):
    r''' Calculate precision for next concept prediction.

    Args:
        predictions:
            Expected shape <batch_size> x <sequence_length> x <vocabulary_size>
        label_ids:
            Expected shape <batch_size> x <sequence_length>
        id2tkn:
            Map from ID to tokens
        id2type:
            Map from ID to token type (e.g. `age`, `cui`, ...)
        token_types (Set[str], optional, defaults to `{'cui'}`:
            On what token types to calculate the Precision. Leave empoty to include all token types.
        prediction_scope:
            How much into the future should we look to accept something as correct:
                - `one` has to be the next concept
                - `age` until the next age token
                - `any` whenever
        shifted_labels:
            Are labels == input_ids, or shifted by one to the left
        predictions_are_scores:
            Are predictions scores for each label_id or really label_ids already
        old_data:
            If set it will load old values for tp/fp/positives/negatives and continue ontop of those
        topk:
            How many predicted labels to consider when calculating precision
        start:
            At what point to start - we will look only at the precision of concepts at positions after start

    Return (Dict[str, ]):
        precision:
            Precision
        tp:
            Number of True positives
        fp:
            Number of False positives
        positives:
            For each label ID a count of positive examples
        negatives
            For each label ID a count of negative examples
    '''
    if predictions_are_scores:
        outputs = np.argsort(-1 * predictions, axis=2)
    else:
        outputs = predictions

    tp = 0
    fp = 0
    positives = {}
    negatives = {}
    # If not shifted_labels label = prediction - 1
    label_position_shift = 0 if shifted_labels else 1
    # If labels are not shifted move the start by one
    start += 0 if shifted_labels else 1

    if old_data:
        tp = old_data['tp']
        fp = old_data['fp']
        positives = old_data['positives']
        negatives = old_data['negatives']

    def prediction_end_index(i, lbl):
        r''' Used below to get the end index for different
        prediction scopes
        '''
        if prediction_scope == 'one':
            return i + 1
        elif prediction_scope == 'any':
            return len(lbl)
        elif prediction_scope == 'age':
            end = len(lbl)
            for j in range(i, len(lbl)):
                if id2type.get(lbl[j], 'unk') == 'age':
                    end = j
                    break
            return end

    for ind, lbl in enumerate(label_ids):
        if start < len(lbl):
            for i in range(start, len(lbl)):
                tkn = str(id2tkn.get(lbl[i], lbl[i]))
                if not token_types or id2type.get(lbl[i], 'unk') in token_types:
                    end = prediction_end_index(i, lbl)

                    # If predictions are scores we can do topk, if not just do simple label match
                    if (predictions_are_scores and any([out in lbl[i:end] for out in outputs[ind][i-label_position_shift][0:topk]])) or \
                        (not predictions_are_scores and outputs[ind][i-label_position_shift] in lbl[i:end]):
                        tp += 1
                        positives[tkn] = positives.get(tkn, 0) + 1
                    else:
                        fp += 1
                        negatives[tkn] = negatives.get(tkn, 0) + 1

    prec = tp / (tp + fp)

    return {'precision': prec, 'tp': tp, 'fp': fp, 'positives': positives, 'negatives': negatives}


def sort_precision_output(precision_output, cdb, main='positives'):
    d = precision_output
    if main == 'positives':
        other = 'negatives'
    else:
        other = 'positives'

    out = sorted([(
        "{:.2f}".format(tp / (tp + d[other].get(cui, 1))), 
        cdb.get_name(cui),
        tp,
        d[other].get(cui, 0),
        cui) for cui, tp in sorted(d[main].items(), key=lambda x: x[1], reverse=True)],
        key=lambda x: x[0], reverse=True)

    out = pd.DataFrame(out, columns=['precision', 'name', main, other, 'cui'])

    return out


class ComputePrecisionHF(object):
    r''' Used for computing precison when working with HF trainer
    '''

    def __init__(self, id2tkn, id2type, batch_size=1000, **kwargs):
        self.id2tkn = id2tkn
        self.id2type = id2type
        self.batch_size = batch_size
        self.kwargs = kwargs

    def __call__(self, p):
        # We will do this in batches, because it can be very memory demanding
        metrics_data = None
        start = 0
        while start < len(p.predictions):
            predictions = np.argmax(p.predictions[start:start+self.batch_size], axis=2) # For speed do not send predictions as scores
            label_ids = p.label_ids[start:start+self.batch_size]
            metrics_data = precision(predictions, label_ids=label_ids, id2tkn=self.id2tkn, id2type=self.id2type, old_data=metrics_data,
                                     predictions_are_scores=False, **self.kwargs)
            start += self.batch_size

        return {
            'precision': metrics_data['precision'],
        }



import logging
import numpy as np
from medcat.utils.matutils import unitvec

def get_embeddings_for_tokens(dataset, cdb, context_type='medium', special_tokens=['<PAD>', '<START>'], normalize=True):
    r''' Given a stream of tokens get the embeddings from MedCAT and make the required maps.

    Args:
        dataset
        cdb
        context_type
        special_tokens
        normalize:
            If True the embedding vectors will be normalized
    Returns:
        embeddings
        tkn2id
        id2tkn
        id2type
        id2type_detailed
    '''
    embeddings = []
    tkn2id = {}
    id2tkn = {}
    id2type = {} # Is the token a CUI/Age/whatever
    id2type_detailed = {} # Will have cdb type_ids for CUIs and same as id2type for rest

    for tkns in dataset['stream']:
        for tkn in tkns:
            tkn = str(tkn)
            if tkn not in tkn2id:
                token_type = 'unk'
                if tkn in cdb.cui2names:
                    token_type = 'cui'
                    # Take the first type_id (should always be the most important one if there are more)
                    token_type_detailed = list(cdb.cui2type_ids.get(cui, ['unk']))[0]

                if tkn in cdb.cui2context_vectors and context_type in cdb.cui2context_vectors[tkn]:
                    vec = cdb.cui2context_vectors[tkn][context_type]
                elif tkn.isdecimal() and int(tkn) >= 0 and int(tkn) <= 300:
                    # Token is an age token
                    token_type = 'age'
                    token_type_detailed = 'age'
                    logging.info("Age token: " + tkn)
                    vec = np.random.rand(300)
                else:
                    logging.info("Token does not have a vector and is not age: " + tkn)
                    vec = np.random.rand(300)

                id2tkn[len(embeddings)] = tkn
                id2type[len(embeddings)] = token_type
                id2type_detailed = token_type_detailed
                tkn2id[tkn] = len(embeddings)

                vec = unitvec(vec) if normalize else vec
                embeddings.append(vec)

    # Add special tokens
    for tkn in special_tokens:
        token_type = 'special'

        id2tkn[len(embeddings)] = tkn
        id2type[len(embeddings)] = token_type
        tkn2id[tkn] = len(embeddings)
        if tkn != '<PAD>':
            embeddings.append(np.random.rand(len(embeddings[0])))
        else:
            embeddings.append(np.zeros(len(embeddings[0])))

    return embeddings, tkn2id, id2tkn, id2type, id2type_detailed


def stream_to_separate_examples(examples):
    r''' Convert a stream to separate examples that can be used to train
    a next concept predictor unable to handle sequences. Use with HF datasets map function.

    '''
    out = {}
    out['input_ids'] = [example[0:i+1] for example in examples['input_ids'] for i in range(len(example) - 1)]
    out['labels'] = [example[i+1] for example in examples['input_ids'] for i in range(len(example) - 1)]
    out['labels_all'] = [example[i+1:] for example in examples['input_ids'] for i in range(len(example) - 1)]
    out['patient_id'] = [id for ind, id in enumerate(examples['patient_id']) for _ in range(len(examples['input_ids'][ind]) - 1)]

    return out



import random
import torch

class SimpleMapTokenizer(object):
    r''' Not even really a tokenizer, will take a list of tokens and
    covert them to IDs

    Args:
        tkn2id
        pad_id
        max_len
        start_id:
            If set it will be prepended to each input example
    '''
    def __init__(self, tkn2id, pad_id, max_len=50, start_id=None, id2type=None, cdb=None):
        self.tkn2id = tkn2id
        self.pad_id = pad_id
        self.max_len = max_len
        self.start_id = start_id
        self.id2type = id2type
        self.cdb = cdb

        # Create id2tkn 
        self.id2tkn = {v:k for k,v in self.tkn2id.items()}


    def __call__(self, text, return_tensors=False):
        out = {'input_ids': [], 'attention_mask': []}

        out['input_ids'] = [self.start_id] + [self.tkn2id[tkn] for tkn in text.split(" ")]
        out['attention_mask'] = [1] * len(out['input_ids'])

        if return_tensors:
            out = {k:torch.tensor([v]) for k,v in out.items()}

        return out


    def decode(self, token_ids, get_names=True):
        tkns = self.convert_ids2tokens(token_ids, get_names=get_names)
        if type(tkns) != list:
            tkns = [tkns]
        return " ".join(tkns)


    def convert_ids2tokens(self, token_ids, get_names=True):
        if type(token_ids) == torch.Tensor:
            token_ids = token_ids.tolist()
        if type(token_ids) == list and type(token_ids[0]) == torch.Tensor:
            token_ids = [x.tolist() for x in token_ids]

        # Same as decode, but needed for compatibility with ecco
        out = []
        if type(token_ids) != list:
            out = [self.id2tkn[int(token_ids)]]
        else:
            # Convert tokens to IDs
            out = [self.id2tkn[int(id)] for id in token_ids]

        if self.cdb is not None and get_names:
            _out = []
            for tkn in out:
                _out.append(self.cdb.get_name(tkn))
                #_out.append(" | ")
            out = _out

        return out



    def tokens_to_ids(self, tokens):
        out = [self.tkn2id[tkn] for tkn in tokens]
        if self.start_id is not None:
            out = [self.start_id] + out
        out = out[:self.max_len]

        return out


    def encode(self, examples):
        r''' Convert 'stream' in the examples from tokens to IDs, save as 'input_ids'. Use with HF datasets.map
        '''
        examples['input_ids'] = [self.tokens_to_ids(stream[:-1]) for stream in examples['stream']]

        return examples


    def augment(self, examples, n_augmentations=1, shuffle_prob=0.5, episode_delimiter='age'):
        r''' Augment the data and add random permutations of examples ontop of existing ones

        Args:
            n_augmentations: Number of augmentations to be done per example
        '''
        if self.id2type is None:
            raise Exception("Requires self.id2type to be set")

        for ind, ids in enumerate(list(examples['input_ids'])):
            for i in range(n_augmentations):
                new_sample = [self.tkn2id['<START>']]
                episode = []
                for id in ids:
                    if len(episode) > 0 and self.id2type[id] == episode_delimiter:
                        if shuffle_prob >= random.random(): random.shuffle(episode)
                        new_sample.extend(episode)

                        # Append also the age
                        new_sample.append(id)
                        # Reset episode
                        episode = []
                    elif id == self.tkn2id['<START>'] or id == self.tkn2id['<PAD>']:
                        # Skip start and pad 
                        pass
                    else:
                        episode.append(id)
                if shuffle_prob >= random.random(): random.shuffle(episode)
                new_sample.extend(episode)

                # Append the new sample to input_ids
                examples['input_ids'].append(new_sample)
                # Append the duplicated patient_id
                examples['patient_id'].append(examples['patient_id'][ind])

        return examples



import setuptools
from setuptools.command.install import install
from setuptools.command.develop import develop
from setuptools.command.egg_info import egg_info

with open("./README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="medflux",
    version="0.1.2",
    author="w-is-h",
    author_email="w.kraljevic@gmail.com",
    description="Temporal modeling of patients and diseases",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/w-is-h/medflux",
    packages=['medflux', 'medflux.datasets', 'medflux.metrics', 'medflux.utils', 'medflux.models'],
    install_requires=[
        'datasets~=1.6.0',
        'ray==1.3.0',

        ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)
